import argparse
from aergia._cli._command._chat import add_chat_subcommand
from aergia._cli._command._image import add_image_subcommand
from aergia._cli._command._models import add_models_subcommand
from aergia._cli._command._session import add_session_subcommand
from rich_argparse import ArgumentDefaultsRichHelpFormatter


def make_parser():
    parser = argparse.ArgumentParser(
        prog="ae[rgia]", formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    parser.add_argument(
        "-b", "--backend", type=str, default="openai", help="select backend type"
    )
    parser.add_argument("--version", action="store_true", help="print the version")
    parser.add_argument("--debug", action="store_true", help="enable debug mode")
    parser.add_argument(
        "--log-level",
        default=None,
        choices=["debug", "info", "warn", "error", "critical"],
        help="configure the log-level of the application",
    )

    subparsers = parser.add_subparsers(dest="subcommand")

    add_chat_subcommand(subparsers)
    # add_tui_subcommand(subparsers)
    # add_repl_subcommand(subparsers)
    # add_code_subcommand(subparsers)
    add_image_subcommand(subparsers)
    # add_execute_subcommand(subparsers)
    add_session_subcommand(subparsers)
    # add_role_subcommand(subparsers)
    # add_info_subcommand(subparsers)
    add_models_subcommand(subparsers)

    return parser



import io
import os
import sys
import httpx
import datetime as dt
from openai import OpenAI
from PIL import Image as PillowImage
from rich.table import Table
from rich_argparse import ArgumentDefaultsRichHelpFormatter
from aergia._model._storage import application_db
from aergia._model._save import save
from aergia._model._load import load_list, load_blob
from aergia._model._data import Image
from aergia._cli._command._utilities import ExitCode


def image(args, stdout, stderr):
    def download(url, file):
        resp = httpx.get(url)
        file.write(resp.content)

    client = OpenAI(api_key=os.environ.get("OPENAI_API_KEY"))
    if not args.prompt:
        prompt = sys.stdin.read()
    else:
        prompt = args.prompt if isinstance(args.prompt, str) else " ".join(args.prompt)

    with stdout.status(
            "Generating image ...", spinner="aesthetic", spinner_style="magenta"
    ):
        response = client.images.generate(prompt=prompt, model=args.model)
        data = response.data[0]
        image_url = data.url
        revised_prompt = data.revised_prompt
        buffer = io.BytesIO()
        download(image_url, buffer)
        created = dt.datetime.fromtimestamp(response.created)
        img = Image(
            id=None,
            name=args.name,
            model=args.model,
            prompt=prompt,
            created=created,
            revised_prompt=revised_prompt,
            blob=buffer.getvalue(),
        )
        db = application_db()
        save(img, db)

    pimg = PillowImage.open(buffer)
    pimg.show(args.name)

    return ExitCode.Success


def list_images(args, stdout, stderr):
    table = Table(title="Generated Images")
    table.add_column("Id", justify="right")
    table.add_column("Name", justify="left", style="green")
    table.add_column("Model", justify="left", style="cyan")
    table.add_column("Created", justify="left", style="yellow")
    table.add_column("Prompt", justify="left", style="white")
    table.add_column("Revised-Prompt", justify="left", style="magenta")

    db = application_db()
    limit = None if args.all else args.limit
    images = load_list(Image, db, limit=limit, offset=args.offset)
    for img in images:
        table.add_row(
            f"{img.id}",
            img.name,
            img.model,
            f"{img.created}",
            img.prompt,
            img.revised_prompt,
        )

    stdout.print(table)

    return ExitCode.Success


def show_image(args, stdout, stderr):
    db = application_db()
    blob = load_blob(Image, key="id", value=args.id, db=db)
    pimg = PillowImage.open(blob)
    pimg.show()
    return ExitCode.Success


def add_image_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        "image",
        help="manage and generate images",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    sub_subparsers = subcommand.add_subparsers(
        dest="session_subcommand", help="Sessions commands"
    )

    generate_subcommand = sub_subparsers.add_parser(
        "generate",
        help="generate images",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )

    generate_subcommand.add_argument(
        "-n",
        "--name",
        type=str,
        default="unnamed",
        help="name or title which will be used for the image",
    )
    generate_subcommand.add_argument(
        "-m",
        "--model",
        default="dall-e-2",
        choices=["dall-e-2", "dall-e-3"],
        help="model to use for generating the image",
    )
    generate_subcommand.add_argument(
        "prompt",
        type=str,
        nargs="*",
        metavar="PROMPT",
        help="prompt which shall be used for generating the image, if None stdin will be read.",
    )
    generate_subcommand.set_defaults(func=image)

    list_subcommand = sub_subparsers.add_parser(
        "list",
        help="list previously generated images",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    list_subcommand.add_argument(
        "-a", "--all", action="store_true", default=False, help="list all images"
    )
    list_subcommand.add_argument(
        "-l", "--limit", type=int, default=10, help="amount of images to list"
    )
    list_subcommand.add_argument(
        "-o", "--offset", type=int, default=0, help="offset for listing images"
    )
    list_subcommand.set_defaults(func=list_images)

    show_subcommand = sub_subparsers.add_parser(
        "show", help="show an image", formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    show_subcommand.add_argument("id", type=int, help="image to show")
    show_subcommand.set_defaults(func=show_image)
    return subcommand



import os
from aergia._cli._command._utilities import ExitCode
from openai import OpenAI
from rich.table import Table
from rich_argparse import ArgumentDefaultsRichHelpFormatter


def models(args, stdout, stderr):
    client = OpenAI(api_key=os.environ.get("OPENAI_API_KEY"))
    models = client.models.list()

    table = Table(title="Available Models of Backend")
    table.add_column("Id/Name", justify="left", style="green")
    table.add_column("Owned By", justify="left", style="cyan")
    table.add_column("Created", justify="left", style="yellow")

    import datetime

    for m in models:
        created = f"{datetime.datetime.fromtimestamp(m.created)}"
        table.add_row(m.id, m.owned_by, created)

    stdout.print(table)

    return ExitCode.Success


def add_models_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        "models",
        help="list available models of the backend",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    subcommand.set_defaults(func=models)
    return subcommand



import asyncio
import sys
import logging

from functools import wraps


from aergia._cli._command._parser import make_parser
from aergia._cli._io import stderr, stdout
from aergia._cli._command._utilities import ExitCode
from aergia._logging import logger
from rich.logging import RichHandler


def _protect(func, *args, **kwargs):
    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            exit_code = func(*args, **kwargs)
        except Exception as ex:
            err_msg = f"Error occurred, details: {ex}"
            logger.error(err_msg)
            stdout.print(err_msg, style="red")
            exit_code = ExitCode.Failure
        except asyncio.CancelledError as ex:
            err_msg = f"Canceled by user, details: {ex}"
            logger.error(err_msg)
            stdout.print(err_msg, style="red")
            exit_code = ExitCode.Failure
        return exit_code

    return wrapper


def main(argv=None):
    parser = make_parser()
    args = parser.parse_args(argv)

    levels = {
        "debug": logging.DEBUG,
        "info": logging.INFO,
        "warn": logging.WARN,
        "error": logging.ERROR,
        "critical": logging.CRITICAL,
    }
    level = args.log_level or "error"
    logging.basicConfig(
        level=levels[level],
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        handlers=[RichHandler(console=stderr, rich_tracebacks=True)],
    )

    if not hasattr(args, "func"):
        parser.error("Subcommand required!")

    # settings = Settings(args)
    app = args.func if args.debug else _protect(args.func)
    exit_code = app(args, stdout, stderr)
    sys.exit(exit_code)


if __name__ == "__main__":
    main()



import sqlite3
from importlib import resources


def initialize(db):
    sql_file = resources.files(__name__) / "init.sql"
    sql_script = sql_file.read_text()
    with sqlite3.connect(db) as con:
        with con as transaction:
            transaction.executescript(sql_script)



from __future__ import annotations
import sqlite3
from pathlib import Path
from aergia._db import initialize


def data_directory(name: str) -> Path:
    return Path.home() / ".local" / "share" / name


def store(name, directory):
    directory.mkdir(exist_ok=True, parents=True)
    db = directory / f"{name}.sqlite"
    with sqlite3.connect(db):
        pass
    return db


def application_db():
    name = "aergia"
    directory = data_directory(name)
    db = store(name, directory)
    initialize(db)
    return db





import argparse
from aergia._cli._command._utilities import default
from rich_argparse import ArgumentDefaultsRichHelpFormatter


def add_code_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        "code", help="Generate code", formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    subcommand.add_argument(
        "text", type=str, nargs=argparse.REMAINDER, help="Code prompt"
    )
    subcommand.add_argument(
        "-r", "--role", type=str, help="Role to be used in addition"
    )
    subcommand.set_defaults(func=default)
    return subcommand



from __future__ import annotations
import datetime
import sqlite3
import io
from inspect import cleandoc
from class_singledispatch import class_singledispatch
from aergia._model._data import Session, Image, Message


@class_singledispatch
def load(type: type[object], key, value, db=None):
    raise TypeError(f"Type {type}, is not supported yet.")


@load.register
def _(image: type[Image], key, value, db):
    with sqlite3.connect(db) as con:
        with con as transaction:
            stmt = cleandoc(f"""
            SELECT id, name, model, prompt, revised_prompt, created, blob FROM images
            WHERE {key} = ?;
            """)
            result = transaction.execute(stmt, (value,))
            row = result.fetchone()
            names = image.model_json_schema()["properties"].keys()
            kwargs = dict(zip(names, row))
            kwargs["created"] = datetime.datetime.fromtimestamp(kwargs["created"])
            return Image(**kwargs)


@load.register
def _(session: type[Session], key, value, db):
    with sqlite3.connect(db) as con:
        with con as transaction:
            stmt = cleandoc(f"""
            SELECT id, name, model, created FROM sessions
            WHERE {key} = ?;
            """)
            result = transaction.execute(stmt, (value,))
            row = result.fetchone()
            if not row:
                raise Exception("Session not found.")
            names = session.model_json_schema()["properties"].keys()
            kwargs = dict(zip(names, row))
            return Session(**kwargs)


@load.register
def _(message: type[Message], key, value, db):
    with sqlite3.connect(db) as con:
        with con as transaction:
            stmt = cleandoc(f"""
            SELECT id, content, model, role, created, session_id FROM messages
            WHERE {key} = ?;
            """)
            result = transaction.execute(stmt, (value,))
            for row in result:
                names = message.model_json_schema()["properties"].keys()
                kwargs = dict(zip(names, row))
                yield Message(**kwargs)


@class_singledispatch
def load_list(type: type[object], db=None, limit=10, offset=0):
    raise TypeError(f"Type {type}, is not supported yet.")


@load_list.register
def _(image: type[Image], db=None, limit=10, offset=0):
    with sqlite3.connect(db) as con:
        if limit is None:
            stmt = "SELECT id, name, model, prompt, revised_prompt, created, blob FROM images;"
            result = con.execute(stmt)
        else:
            stmt = cleandoc("""
                SELECT id, name, model, prompt, revised_prompt, created, blob FROM images
                ORDER BY id
                LIMIT ? OFFSET ?;
            """)
            result = con.execute(stmt, (limit, offset))
        names = Image.model_json_schema()["properties"].keys()
        rows = result.fetchall()
        images = []
        for row in rows:
            kwargs = dict(zip(names, row))
            kwargs["created"] = datetime.datetime.fromtimestamp(kwargs["created"])
            images.append(Image(**kwargs))
        return images


@load_list.register
def _(image: type[Session], db=None, limit=10, offset=0):
    with sqlite3.connect(db) as con:
        if limit is None:
            stmt = "SELECT id, name, model, created FROM sessions;"
            result = con.execute(stmt)
        else:
            stmt = cleandoc("""
                SELECT id, name, model, created FROM sessions
                ORDER BY id
                LIMIT ? OFFSET ?;
            """)
            result = con.execute(stmt, (limit, offset))
        names = Session.model_json_schema()["properties"].keys()
        rows = result.fetchall()
        sessions = []
        for row in rows:
            kwargs = dict(zip(names, row))
            sessions.append(Session(**kwargs))
        return sessions


@class_singledispatch
def load_blob(type: type[object], id, db=None):
    raise TypeError(f"Type {type}, is not supported yet.")


@load_blob.register
def _(image: type[Image], key, value, db=None):
    with sqlite3.connect(db) as con:
        stmt = f"SELECT blob FROM images WHERE {key} = ?;"
        result = con.execute(stmt, (value,))
        row = result.fetchone()
        return io.BytesIO(row[0])



import io
from enum import IntEnum
from inspect import cleandoc


class ExitCode(IntEnum):
    Success = 0
    Failure = 1


class TextBuffer:
    """Class to manage a text buffer using StringIO."""

    def __init__(self):
        """
        Initializes a new instance of the TextBuffer class.
        """
        self._buffer = io.StringIO()

    def __enter__(self):
        """
        Enters the runtime context related to this object.

        Returns:
            The TextBuffer instance.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exits the runtime context related to this object.

        Args:
            exc_type: The exception type.
            exc_val: The exception value.
            exc_tb: The traceback object.
        """
        self._buffer.close()

    def append(self, text):
        """
        Appends text to the buffer.

        Args:
            text: The text to append to the buffer.
        """
        self._buffer.write(text)

    def __iadd__(self, other):
        """
        Implements the in-place addition operation.

        Args:
            other: The text to add to the buffer.

        Returns:
            The TextBuffer instance.
        """
        self._buffer.write(other)
        return self

    @property
    def content(self):
        """
        Retrieves the content of the buffer.

        Returns:
            The content of the buffer as a string.
        """
        return self._buffer.getvalue()


def default(args, stdout, stderr):
    err_msg = cleandoc(f"""
    This function is not implemented yet!

        Details:
            Args: {args}

        In case you need more details, consider running the command with the `--debug` flag.
    """)
    raise NotImplementedError(err_msg)



from aergia._cli._command._utilities import default
from rich_argparse import ArgumentDefaultsRichHelpFormatter


def add_info_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        subparsers,
        "info",
        help="Print info about data store and config file, etc.",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    subcommand.set_defaults(func=default)
    return subcommand



from __future__ import annotations
import os
from pathlib import Path
import tomli
from collections import ChainMap
from inspect import cleandoc


def default_config():
    DEFAULT_CONFIG = cleandoc("""
    [defaults]
    db_name = "aergia"  # base name of the database which will be used to store chats, images, etc.
    backend="openai"    # backend which should be used by default

    [backends.openai]
    name = "openai"                         # name of the backend
    base_url = "https://api.openai.com/v1"  # base url of the backend api
    token=""                                # token used to authenticate against the backend

    [commands.image]
    generate.model = "dall-e-3" # default model to use
    generate.name = "unnamed"   # default name to use if none is specified
    list.all = false            # should all images be listed if true, limit will be ignored
    list.limit = 10             # amount of images which should be listed by default
    list.offset = 0             # offset from which the images should be started to be listed

    [commands.chat]
    model = "gpt-4o"
    # role = None               # role which should be used by default (if set session will be ignoed)
    # session = None            # session which should be used by default (conflicts with role)
    # context = None            # context file which should be passed as context

    # [commands.models]
    # currently does not have any associated settings
    """)
    config = tomli.loads(DEFAULT_CONFIG)
    config["commands"]["chat"]["role"] = None
    config["commands"]["chat"]["session"] = None
    config["commands"]["chat"]["context"] = None
    config["commands"]["models"] = {}
    return config


def load_config_file(path: Path):
    with open(path, "rb") as f:
        return tomli.load(f)


def user_config(cfg: Path):
    if not cfg.exists():
        return {}
    return load_config_file(cfg)


def env_config(env):
    def commands_env():
        cfg = {}
        return cfg

    def defaults():
        cfg = {}
        return cfg

    def openai_env():
        cfg = {}
        settings = {
            "OPENAI_API_KEY": "token",
            "OPENAI_BASE_URL": "base_url",
        }
        settings_in_env = {k: v for k, v in settings.items() if k in env}
        for env_name, cfg_name in settings_in_env.items():
            cfg[cfg_name] = env[env_name]

        if len(cfg) > 0:
            cfg["name"] = "openai"
        return cfg

    config = {
        "defaults": {"defaults": defaults()},
        "backends": {"openai": openai_env()},
        "commands": {"commands": commands_env()},
    }

    return config


def command_line(arguments):
    command = arguments.subcommand
    config = {"defaults": {"db_name": "aergia", "backend": "openai"}}

    def chat_args(args):
        return {"chat": {"model": args.model}}

    def image_args(args):
        match args.subcommand:
            case "generate":
                return {"model": args.model, "name": args.name}
            case "list":
                return {"all": args.all, "limit": args.limit, "offset": args.offset}
            case _:
                return {}

    dispatcher = {
        "chat": chat_args,
        "image": image_args,
    }
    try:
        config[command] = dispatcher[command](arguments)
    except KeyError:
        config[command] = {}

    return config


class Settings:
    def __init__(self, args, env=None, cfg=None, defaults=None):
        env = env or os.environ.copy()
        cfg = cfg or (Path.home() / "aergia.toml")
        self._settings = ChainMap(
            command_line(args),
            env_config(env),
            user_config(cfg),
            defaults or default_config(),
        )

    def __getattr__(self, name):
        try:
            return self._settings[name]
        except KeyError:
            raise AttributeError(f"'Settings' object has no attribute '{name}'")



from openai import OpenAI, AsyncClient
from enum import Enum, auto


class Type(Enum):
    Async = auto()
    Sync = auto()


def _select_client(backend, client_type):
    OPENAI = {Type.Async: AsyncClient, Type.Sync: OpenAI}
    backends = {"openai": OPENAI}

    try:
        selected_backend = backends[backend]
    except KeyError as ex:
        err_msg = f"Backend ({backend}) is not supported."
        raise Exception(err_msg) from ex

    try:
        klass = selected_backend[client_type]
    except KeyError as ex:
        err_msg = f"Streaming is not supported by {selected_backend}."
        raise Exception(err_msg) from ex

    return klass


def build_client(backend, api_key, base_url=None, client_type=Type.Sync):
    if base_url is None:
        base_url = "https://api.openai.com/v1"
    Client = _select_client(backend, client_type)
    return Client(base_url=base_url, api_key=api_key)



from __future__ import annotations
import sqlite3
from inspect import cleandoc
from functools import singledispatch

from aergia._model._data import Session, Image, Message


@singledispatch
def save(type, db=None):
    raise TypeError(f"Type {type.__class__}, is not supported yet.")


@save.register
def _(image: Image, db=None):
    with sqlite3.connect(db) as con:
        with con as transaction:
            stmt = cleandoc("""
            INSERT INTO images (name, model, prompt, revised_prompt, created, blob)
            VALUES (?, ?, ?, ?, ?, ?);
            """)
            result = transaction.execute(
                stmt,
                (
                    image.name,
                    image.model,
                    image.prompt,
                    image.revised_prompt or "<NO-REVISED-PROMPT>",
                    image.created.timestamp(),
                    image.blob,
                ),
            )
        image.id = result.lastrowid
        return image


@save.register
def _(session: Session, db=None):
    with sqlite3.connect(db) as con:
        with con as transaction:
            stmt = cleandoc("""
            INSERT INTO sessions (name, model, created)
            VALUES (?, ?, ?);
            """)
            result = transaction.execute(
                stmt,
                (session.name, session.model, session.created),
            )
        session.id = result.lastrowid
        return session


@save.register
def _(message: Message, db=None):
    with sqlite3.connect(db) as con:
        with con as transaction:
            stmt = cleandoc("""
            INSERT INTO messages (content, model, role, created, session_id)
            VALUES (?, ?, ?, ?, ?);
            """)
            result = transaction.execute(
                stmt,
                (message.content, message.model, message.role, message.created, message.session_id),
            )
        message.id = result.lastrowid
        return message



import os


def env(obj):
    return (obj.__to_envkey__(), obj.__to_str__())


def env_key(obj):
    return obj.__to_envkey__()


def from_env(type, name, env=None):
    def boolean(v):
        true = ["yes", "true", "on", "enabled"]
        false = ["no", "false", "off", "disabled"]
        v = v.lower()
        if v in true:
            return True
        if v in false:
            return False

        raise ValueError(f"Connot convert value: {v}, to bool.")

    converts = {int: int, float: float, str: str, bool: boolean}
    if type not in converts:
        converter = getattr(type, "__from_str__", str)
    else:
        converter = converts[type]

    env = env or os.environ
    if name not in env:
        return None

    return converter(env[name])


def normalize(identifier):
    identifier = identifier.replace("-", "_")
    identifier = identifier.upper()
    return identifier



import argparse
from aergia._cli._command._utilities import default
from rich_argparse import ArgumentDefaultsRichHelpFormatter


def add_execute_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        "execute",
        help="Generate and execute a command",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    subcommand.add_argument(
        "text", type=str, nargs=argparse.REMAINDER, help="Command prompt"
    )
    subcommand.add_argument(
        "-s", "--shell", type=str, help="Specify the shell to be used for execution"
    )
    subcommand.add_argument(
        "-r", "--role", type=str, help="Role to be taken into account"
    )
    subcommand.set_defaults(func=default)
    return subcommand



import os
import argparse
import uuid
from rich.markdown import Markdown
from rich_argparse import ArgumentDefaultsRichHelpFormatter
from aergia._cli._command._utilities import TextBuffer, ExitCode
from aergia._client import build_client, Type
from aergia._model._save import save
from aergia._model._load import load
from aergia._model._data import Session, Message
from aergia._model._storage import application_db


def chat(args, stdout, stderr):
    def is_chunk_valid(chunk):
        return chunk.choices is not None and len(chunk.choices) > 0

    def sync_chat(client, model, messages):
        response = client.chat.completions.create(
            model=model,
            messages=messages,
            stream=True,
        )
        with stdout.status(
            "Processing request ...", spinner="aesthetic", spinner_style="cyan"
        ):
            with TextBuffer() as buffer:
                stream = (c for c in response if is_chunk_valid(c))
                for chunk in stream:
                    text = chunk.choices[0].delta.content or ""
                    model = chunk.model
                    buffer.append(text)
                return buffer.content, model

    def create_session(name=None):
        name = name or f"Temp-{uuid.uuid4()}"
        try:
            session = load(Session, key="name", value=args.session, db=db)
        except Exception:
            session = Session.new(name, args.model)
            save(session, db)
        return session

    api_key = os.environ.get("OPENAI_API_KEY")
    client = build_client(backend="openai", api_key=api_key, client_type=Type.Sync)
    db = application_db()
    session = create_session(args.session)
    context = "" if not args.context else args.context.read()
    content = " ".join(args.text) + context

    user_msg = Message.user(content, session.id)
    save(user_msg, db)

    messages = list(load(Message, key="session_id", value=session.id, db=db))
    messages = [{"role": m.role, "content": m.content} for m in messages]
    messages.append({"role": user_msg.role, "content": user_msg.content})

    content, model = sync_chat(client, model=args.model, messages=messages)

    assistant_msg = Message.assistant(content, model, session.id)
    save(assistant_msg, db)

    stdout.print(Markdown(assistant_msg.content))
    return ExitCode.Success


def add_chat_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        "chat",
        help="chat with the ai system",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    subcommand.add_argument(
        "text", type=str, nargs=argparse.REMAINDER, help="Text input"
    )
    subcommand.add_argument("-r", "--role", type=str, help="Choose a role")
    subcommand.add_argument(
        "-s", "--session", nargs="?", const=True, help="Create or reuse a session"
    )
    subcommand.add_argument(
        "-c", "--context", type=argparse.FileType("r"), help="File or stdin (-)"
    )
    subcommand.add_argument(
        "-m",
        "--model",
        default="gpt-4o",
        choices=["gpt-4o", "gpt-4", "gpt-4-turbo", "gpt-3.5-turbo"],
        help="Select a model",
    )
    subcommand.set_defaults(func=chat)
    return subcommand



from typing import Generic, TypeVar
from dataclasses import dataclass
from aergia._config import _env

T = TypeVar("T")


@dataclass(frozen=True)
class Setting(Generic[T]):
    name: str
    value: T | None = None
    prefix: str | None = None
    type: T = T
    default: T | None = None
    description: str = ""

    def __to_envkey__(self):
        name = f"{self.prefix}-" if self.prefix is not None else ""
        name += self.name
        name = _env.normalize(name)
        return f"{name}"

    def __to_str__(self):
        value = self.value if self.value is not None else self.default
        if value is None:
            return ""
        return f"{value}"

    def __from_str__(self, string):
        return self.type(string)


class Namespace:
    name: str
    settings: list
    description: str = ""



import argparse

from rich.table import Table
from rich.markdown import Markdown
from rich_argparse import ArgumentDefaultsRichHelpFormatter

from aergia._cli._command._utilities import ExitCode, default
from aergia._model._data import Session, Message
from aergia._model._load import load_list, load
from aergia._model._storage import application_db


def list_sessions(args, stdout, stderr):
    table = Table(title="Chat Sessions")
    table.add_column("Id", justify="right")
    table.add_column("Name", justify="left", style="green")
    table.add_column("Model", justify="left", style="cyan")
    table.add_column("Created", justify="left", style="yellow")

    db = application_db()
    limit = None if args.all else args.limit
    sessions = load_list(Session, db, limit=limit, offset=args.offset)
    for s in sessions:
        table.add_row(
            f"{s.id}",
            s.name,
            s.model,
            f"{s.created}",
        )

    stdout.print(table)

    return ExitCode.Success


def show_session(args, stdout, stderr):
    from rich import print
    from rich.panel import Panel
    db = application_db()
    messages = list(load(Message, key="session_id", value=args.id, db=db))
    for m in messages:
        color = "blue" if m.role == 'user' else 'magenta'
        p = Panel(Markdown(m.content), title=m.role, style=color)
        stdout.print(p)

    return ExitCode.Success


def add_session_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        "session",
        help="manage sessions",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )

    sub_subparsers = subcommand.add_subparsers(
        dest="session_subcommand", help="Sessions commands"
    )

    list_command = sub_subparsers.add_parser(
        "list",
        help="list all sessions",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    list_command.add_argument(
        "-a", "--all", action="store_true", default=False, help="list all images"
    )
    list_command.add_argument(
        "-l", "--limit", type=int, default=10, help="amount of images to list"
    )
    list_command.add_argument(
        "-o", "--offset", type=int, default=0, help="offset for listing images"
    )
    list_command.set_defaults(func=list_sessions)

    show_command = sub_subparsers.add_parser(
        "show",
        help="show contents of a session",
        formatter_class=ArgumentDefaultsRichHelpFormatter,
    )
    show_command.add_argument("id", type=int, help="session id of session to show")
    show_command.set_defaults(func=show_session)

    return subcommand



def toml_key(obj):
    return obj.__to_envkey__()


def from_toml(type, name, path):
    pass


def normalize(identifier):
    identifier = identifier.replace("-", "_")
    identifier = identifier.upper()
    return identifier



from aergia._cli._command._utilities import default
from rich_argparse import ArgumentDefaultsRichHelpFormatter


def add_role_subcommand(subparsers):
    subcommand = subparsers.add_parser(
        "role", help="Manage roles", formatter_class=ArgumentDefaultsRichHelpFormatter
    )
    subcommand.set_defaults(func=default)

    sub_subparsers = subcommand.add_subparsers(
        dest="role_subcommand", help="Roles commands"
    )

    list_command = sub_subparsers.add_parser("list", help="List all roles")

    show_command = sub_subparsers.add_parser("show", help="Show a role")
    show_command.add_argument("name", type=str, help="Role name")

    create_command = sub_subparsers.add_parser("create", help="Create a role")
    create_command.add_argument("name", type=str, help="Role name")
    create_command.add_argument("text", type=str, nargs="?", help="Role description")

    update_command = sub_subparsers.add_parser("update", help="Update a role")
    update_command.add_argument("name", type=str, help="Role name")
    update_command.add_argument(
        "text", type=str, nargs="?", help="New role description"
    )

    delete_command = sub_subparsers.add_parser("delete", help="Delete a role")
    delete_command.add_argument("name", type=str, help="Role name")

    alias_command = sub_subparsers.add_parser(
        "alias", help="Create an alias for a role"
    )
    alias_command.add_argument("role", type=str, help="Role name")
    alias_command.add_argument("alias", type=str, help="Alias name")
    return subcommand



from __future__ import annotations

import datetime as dt
from pydantic import BaseModel


class Session(BaseModel):
    id: int | None
    name: str
    model: str
    created: dt.datetime

    @classmethod
    def new(cls, name, model):
        return cls(
            id=None,
            name=name,
            model=model,
            created=dt.datetime.now(),
        )


class Message(BaseModel):
    id: int | None
    content: str
    model: str
    role: str
    created: dt.datetime
    session_id: int | None

    @classmethod
    def user(cls, content, session_id=None):
        return cls(
            id=None,
            content=content,
            model="",
            role="user",
            created=dt.datetime.now(),
            session_id=session_id,
        )

    @classmethod
    def assistant(cls, content, model, session_id=None):
        return cls(
            id=None,
            content=content,
            model=model,
            role="assistant",
            created=dt.datetime.now(),
            session_id=session_id,
        )


class Image(BaseModel):
    id: int | None
    name: str
    model: str
    prompt: str
    revised_prompt: str | None
    created: dt.datetime
    blob: bytes
